from mux_demux_ws.secondary_base_consumer import SecondaryDefaultConsumer


class AlertConsumer(SecondaryDefaultConsumer):

    @property
    def default_group_name(self):
        return "alerts_{}".format(self.user.id)



from rest_framework import permissions
from rest_framework.permissions import IsAdminUser, SAFE_METHODS, BasePermission


class CustomObjectPermissions(permissions.DjangoObjectPermissions):
    """
    Similar to `DjangoObjectPermissions`, but adding 'view' permissions.
    """
    perms_map = {
        'GET': ['%(app_label)s.view_%(model_name)s'],
        'OPTIONS': ['%(app_label)s.view_%(model_name)s'],
        'HEAD': ['%(app_label)s.view_%(model_name)s'],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }

    def has_permission(self, request, view):
        return request.user and request.user.is_authenticated



# todo: can bitwise or these bad boys
class IsAdminUserOrReadOnly(IsAdminUser):
    def has_permission(self, request, view):
        is_admin = super(
            IsAdminUserOrReadOnly,
            self).has_permission(request, view)
        # Python3: is_admin = super().has_permissions(request, view)
        return request.method in SAFE_METHODS or is_admin

class ReadOnly(BasePermission):
    def has_permission(self, request, view):
        return request.method in SAFE_METHODS

class IsEmailVerified(BasePermission):
    def has_permission(self, request, view):
        user = request.user
        email_address = user.emailaddress_set.get(email=user.email)
        return email_address.verified






from collections import OrderedDict

from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response


class CustomPagination(PageNumberPagination):
    page_size_query_param = 'page_size'
    def get_paginated_response(self, data):
        return Response(OrderedDict([
            ('last_page', self.page.paginator.count),
            ('count', self.page.paginator.count),
            ('current', self.page.number),
            ('next', self.get_next_link()),
            ('previous', self.get_previous_link()),
            ('results', data)
        ]))
    def paginate_queryset(self, queryset, request, view=None):
        if 'no_page' in request.query_params:
            return None
        return super().paginate_queryset(queryset, request, view)


# When the _meta object was formalized, this exception was moved to
# django.core.exceptions. It is retained here for backwards compatibility
# purposes.
# from primary_key import make_id

import json

from django.core import validators
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator, MaxValueValidator
# When the _meta object was formalized, this exception was moved to
# django.core.exceptions. It is retained here for backwards compatibility
# purposes.
from django.db import connection
from django.db import models
from django.db.models import IntegerField
from django.utils.functional import cached_property
from rest_framework import serializers


class StringArrayField(models.CharField):
    description = "A string array field that automatically dumps and loads between json and string for db storage."

    def get_db_prep_value(self, value, *args, **kwargs):
        if value is None:
            return None
        return json.dumps(value)

    def to_python(self, value):
        try:
            return json.loads(value)
        except Exception as e:
            raise ValidationError(e)

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)


class BigIntStrReprSerializerField(serializers.IntegerField):
    def to_internal_value(self, data):
        if isinstance(data, str) and len(data) > self.MAX_STRING_LENGTH:
            self.fail('max_string_length')

        try:
            data = self.re_decimal.sub('', str(data))
        except (ValueError, TypeError):
            self.fail('invalid')
        return data

    def to_representation(self, value):
        if value is None:
            return None
        return str(value)

class JsonFieldSerializerField(serializers.Field):

    def to_representation(self, value):
        return value

    def to_internal_value(self, data):
        return data

class IntStrMinValueValidator(MinValueValidator):
    def compare(self, a, b):
        return int(a) < int(b)


class IntStrMaxValueValidator(MaxValueValidator):
    def compare(self, a, b):
        return int(a) > int(b)

class BigIntStrRepr(models.BigIntegerField):
    description = "A Big Integer field that gets converted to a string for python / serialization use."

    @cached_property
    def validators(self):
        # These validators can't be added at field initialization time since
        # they're based on values retrieved from `connection`.
        validators_ = super(IntegerField, self).validators
        internal_type = self.get_internal_type()
        min_value, max_value = connection.ops.integer_field_range(internal_type)
        if min_value is not None and not any(
                (
                        isinstance(validator, validators.MinValueValidator) and (
                        validator.limit_value()
                        if callable(validator.limit_value)
                        else validator.limit_value
                ) >= min_value
                ) for validator in validators_
        ):
            validators_.append(IntStrMinValueValidator(min_value))
        if max_value is not None and not any(
                (
                        isinstance(validator, validators.MaxValueValidator) and (
                        validator.limit_value()
                        if callable(validator.limit_value)
                        else validator.limit_value
                ) <= max_value
                ) for validator in validators_
        ):
            validators_.append(IntStrMaxValueValidator(max_value))
        return validators_

    def get_db_prep_value(self, value, *args, **kwargs):
        if value is None:
            return None
        elif isinstance(value, str):
            try:
                return int(value)
            except Exception as e:
                raise ValidationError(e)
        else:
            return value

    def to_python(self, value):
        if value is None:
            return None
        try:
            return str(value)
        except Exception as e:
            raise ValidationError(e)

    def from_db_value(self, value, expression, connection, context):
        return self.to_python(value)

    def to_representation(self, value):
        if value is None:
            return None
        return str(value)




from rest_framework.filters import SearchFilter, OrderingFilter, BaseFilterBackend

class UserFilterForgiving(BaseFilterBackend):
    """
    A filter backend that limits results to those where they are the foreign key 'user'
    """

    def filter_queryset(self, request, queryset, view):
        try:
            return queryset.filter(user=request.user)
        except:
            return queryset

class WithCreatedSearchFilter(SearchFilter):

    def get_search_terms(self, request):
        original_terms = super().get_search_terms(request)
        if "created" in original_terms:
            original_terms.remove("created")
            original_terms.append("id")
        return original_terms

class WithCreatedOrderingFilter(OrderingFilter):

    def get_ordering(self, request, queryset, view):
        params = request.query_params.get(self.ordering_param)
        if params:
            fields = [param.strip() for param in params.split(',')]
            if "created" in fields:
                fields[fields.index("created")] = "id"
            if "-created" in fields:
                fields[fields.index("-created")] = "-id"
            ordering = self.remove_invalid_fields(queryset, fields, view, request)
            if ordering:
                return ordering


class OnTheFlySearchFilter(WithCreatedSearchFilter):

    def filter_queryset(self, request, queryset, search_fields=None, serializer_class=None):

        class DummyView:
            pass

        if search_fields:
            setattr(DummyView, "search_fields", search_fields)

        if serializer_class:
            setattr(DummyView, "get_serializer_class", lambda self: serializer_class)

        return super().filter_queryset(request, queryset, DummyView)

class OnTheFlyOrderingFilter(WithCreatedOrderingFilter):

    def filter_queryset(self, request, queryset, ordering_fields=None, ordering=None, serializer_class=None):

        class DummyView:
            pass

        if ordering_fields:
            setattr(DummyView, "ordering_fields", ordering_fields)

        if ordering:
            setattr(DummyView, "ordering", ordering)

        if serializer_class:
            setattr(DummyView, "get_serializer_class", lambda self: serializer_class)

        return super().filter_queryset(request, queryset, DummyView)



from functools import wraps

from django_utils.fields import BigIntStrRepr, BigIntStrReprSerializerField, JsonFieldSerializerField
from django_utils.mock_request import MockRequest
from global_requests import get_thread_user
from guardian_queryset.serializers import ValidateForeignKeyUserMixin, AssignActiveUserDefaultPermissionMixin
from rest_framework import serializers
from reversable_primary_key.serializers import CreatedSerializerMixin
from jsonfield import JSONField


class BigIntStrReprSerializerMixin(serializers.Serializer):

    @property
    def serializer_field_mapping(self):
        mapping = super().serializer_field_mapping.copy()
        mapping[BigIntStrRepr] = BigIntStrReprSerializerField
        return mapping

class JsonFieldSerializerMixin(serializers.Serializer):
    @property
    def serializer_field_mapping(self):
        mapping = super().serializer_field_mapping.copy()
        mapping[JSONField] = JsonFieldSerializerField
        return mapping

class WithMockRequesterSerializerMixin(serializers.Serializer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "request" not in self._context:
            user = get_thread_user()
            self._context["request"] = MockRequest(user)


class RequireAllFieldsSerializerMixin(serializers.Serializer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        not_required_fields = []
        if hasattr(self.Meta, "not_required_fields"):
            not_required_fields = self.Meta.not_required_fields
        if not hasattr(self.Meta, "extra_kwargs"):
            self.Meta.extra_kwargs = {}
        fields_dict = self.get_fields()
        for name in fields_dict:
            value = fields_dict[name]
            if name not in not_required_fields and not value.read_only:
                if name not in self.Meta.extra_kwargs:
                    self.Meta.extra_kwargs[name] = {}
                self.Meta.extra_kwargs[name]["required"] = True
                self.Meta.extra_kwargs[name]["allow_null"] = False

class AddUserMixin(ValidateForeignKeyUserMixin):
    def validate(self, data):
        data = super(AddUserMixin, self).validate(data)
        current_user = self.context['request'].user
        return {
            **data,
            'user': current_user
        }

class IsValidExceptionSerializerMixin(serializers.Serializer):

    def is_valid(self, **kwargs):
        return super().is_valid(**{
            'raise_exception': True,
            **kwargs,
        })

class DefaultPrivateModelSerializer(AddUserMixin,
                                    AssignActiveUserDefaultPermissionMixin,
                                    IsValidExceptionSerializerMixin,
                                    CreatedSerializerMixin,
                                    ValidateForeignKeyUserMixin,
                                    BigIntStrReprSerializerMixin,
                                    WithMockRequesterSerializerMixin,
                                    serializers.ModelSerializer):
    pass

def with_serializer(
        fields=None,
        exclude=None,
        model=None,
        base=DefaultPrivateModelSerializer,
        serializer_class=None,
        meta={},
        serializer_fields = {},
        many_write=False,
        update=False,
        pass_only=False,
        validate=True,
):

    class OnTheFlySerializer(base):
        class Meta:
            pass
        setattr(Meta, "model", model)
        if fields:
            setattr(Meta, "fields", fields)
        if exclude:
            setattr(Meta, "exclude", exclude)

        for k in meta.keys():
            setattr(Meta, k, meta[k])

    for k in serializer_fields.keys():
        setattr(OnTheFlySerializer, k, serializer_fields[k])

    effective_serializer_cls = serializer_class or OnTheFlySerializer
    def decorator(func):
        @wraps(func)
        def wrapped_func(self, request, *args, **kwargs):
            if request.method == 'GET' or pass_only:
                return func(self, request, *args, serializer_class=effective_serializer_cls, **kwargs)
            elif update:
                context = self.get_serializer_context()
                class UpdateSerializer(effective_serializer_cls):
                    def __init__(self, instance, *args, **kwargs):
                        super().__init__(instance, *args, data=request.data, context=context, **kwargs)
                # def do_serialized_update(instance, *args, **kwargs):
                #     serializer =effective_serializer_cls(instance, *args, data=request.data, context=self.get_serializer_context(), **kwargs)
                #     serializer.is_valid()
                #     serializer.save()
                return func(self, request, *args, **kwargs, serializer_class=UpdateSerializer)
            else:
                serializer = effective_serializer_cls(data=request.data, context=self.get_serializer_context(), many=many_write)
                validate and serializer.is_valid()
                return func(self, request, *args, serializer=serializer, **kwargs)
        return wrapped_func
    return decorator

class ReadWriteSerializerMixin(object):
    """
    Overrides get_serializer_class to choose the read serializer
    for GET requests and the write serializer for POST requests.

    Set read_serializer_class and write_serializer_class attributes on a
    viewset.
    """

    read_serializer_class = None
    write_serializer_class = None

    def get_serializer_class(self):
        if self.action in ["create", "update", "partial_update", "destroy"]:
            return self.get_write_serializer_class()
        return self.get_read_serializer_class()

    def get_read_serializer_class(self):
        assert self.read_serializer_class is not None, (
                "'%s' should either include a `read_serializer_class` attribute,"
                "or override the `get_read_serializer_class()` method."
                % self.__class__.__name__
        )
        return self.read_serializer_class

    def get_write_serializer_class(self):
        assert self.write_serializer_class is not None, (
                "'%s' should either include a `write_serializer_class` attribute,"
                "or override the `get_write_serializer_class()` method."
                % self.__class__.__name__
        )
        return self.write_serializer_class


from pprint import pprint

from rest_framework.views import exception_handler


def custom_exception_handler(exc, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = exception_handler(exc, context)

    # Now add the HTTP status code to the response.
    if response is not None:
        response.data['status_code'] = response.status_code
        pprint(response.data)

    return response


from functools import wraps

from django.db import models
from django.db.models import QuerySet
from django.http import HttpResponse
from django_utils.filters import OnTheFlyOrderingFilter, OnTheFlySearchFilter, WithCreatedSearchFilter, \
    WithCreatedOrderingFilter
from django_utils.permissions import CustomObjectPermissions
from guardian_queryset.filters import GuardianViewPermissionsFilter
from rest_framework import mixins, generics
from rest_framework import status
from rest_framework import viewsets, filters
from rest_framework.response import Response
from rest_framework.viewsets import GenericViewSet


class DefaultPrivateApiViewMixin(generics.GenericAPIView):
    filter_backends = [WithCreatedOrderingFilter,WithCreatedSearchFilter,GuardianViewPermissionsFilter, filters.OrderingFilter]
    permission_classes = [CustomObjectPermissions]
    ordering = ['-id']


class DefaultPrivateViewSetMixin(DefaultPrivateApiViewMixin):


    def get_defaults(self):
        return {'user': self.request.user}

    def get_queryset(self):
        return self.queryset.filter(**self.get_defaults())

    def perform_create(self, serializer):
        serializer.save(**self.get_defaults())

class PaginatedListViewSetMixin(viewsets.GenericViewSet):

    def get_paginated_list_response(self, queryset, serializer_class=None):
        page = self.paginate_queryset(queryset)
        if page is not None:
            effective_queryset = page
        else:
            effective_queryset = queryset
        if serializer_class:
            context = self.get_serializer_context()
            serializer = serializer_class(effective_queryset, many=True, context=context)
        else:
            serializer = self.get_serializer(effective_queryset, many=True)
        return self.get_paginated_response(serializer.data)


def as_response(
        paginate=True,
        many=True,
        filter=True,
        filter_backends=None,
        filter_all_backends=False,
        order_fields=None,
        ordering=None,
        search_fields=None,
        serializer_class=None,
        status=status.HTTP_200_OK
):

    def decorator(func):
        _serializer_class = serializer_class
        @wraps(func)
        def wrapped_func(self, request, *args, **kwargs):
            serializer_class=kwargs["serializer_class"] if "serializer_class" in kwargs else None
            try:
                res = func(self, request, *args, **kwargs)
            except TypeError:
                kwargs.pop("serializer_class")
                res = func(self, request, *args, **kwargs)
            if isinstance(res, QuerySet) or isinstance(res, models.Model):
                querysetOrObj = res
            elif isinstance(res, HttpResponse):
                return res
            elif res is None:
                return Response(status=status)
            else:
                raise ValueError("Return value must be a QuerySet, HttpResponse or None")
            effective_serializer_class = _serializer_class or serializer_class or self.get_serializer_class()
            if isinstance(querysetOrObj, QuerySet):
                if filter_backends:
                    for backend in list(filter_backends):
                        querysetOrObj = backend().filter_queryset(self.request, querysetOrObj, self)
                    if filter_all_backends:
                        querysetOrObj = self.filter_queryset(querysetOrObj)
                elif filter:
                    querysetOrObj = self.filter_queryset(querysetOrObj)
                if ordering:
                    querysetOrObj = OnTheFlyOrderingFilter().filter_queryset(request, querysetOrObj, order_fields, ordering, effective_serializer_class)
                if search_fields:
                    querysetOrObj = OnTheFlySearchFilter().filter_queryset(request, querysetOrObj, search_fields, effective_serializer_class)
                if paginate:
                    return PaginatedListViewSetMixin.get_paginated_list_response(self, querysetOrObj, effective_serializer_class)
            context = self.get_serializer_context()
            serializer = effective_serializer_class(querysetOrObj, many=many, context=context)
            return Response(serializer.data)
        return wrapped_func
    return decorator

class NonUpdateModelViewset(
    mixins.CreateModelMixin,
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    GenericViewSet):
    pass



import traceback
from urllib.parse import parse_qs

from channels.auth import AuthMiddlewareStack
from django.contrib.auth.models import AnonymousUser
from django.db import close_old_connections
from django.utils.translation import gettext_lazy as _
from rest_framework.authentication import TokenAuthentication
from rest_framework_simplejwt.authentication import JWTAuthentication as BaseJSONWebTokenAuthentication
from rest_framework_simplejwt.exceptions import TokenError, InvalidToken
from rest_framework_simplejwt.settings import api_settings


# TODO: deprecate
class TokenAuthSupportQueryString(TokenAuthentication):
    """
    Extend the TokenAuthentication class to support querystring authentication
    in the form of "http://www.example.com/?auth_token=<token_key>"
    """
    def authenticate(self, request):
        # Check if 'token_auth' is in the request query params.
        # Give precedence to 'Authorization' header.
        # print(request.query_params['auth_token'])
        if 'auth_token' in request.query_params and \
                'HTTP_AUTHORIZATION' not in request.META:
            return self.authenticate_credentials(request.query_params.get('auth_token'))
        else:
            return super(TokenAuthSupportQueryString, self).authenticate(request)



class WsTokenAuthMiddleware:
    """
    Token authorization middleware for Django Channels 2
    see:
    https://channels.readthedocs.io/en/latest/topics/authentication.html#custom-authentication
    """

    def __init__(self, inner):
        self.inner = inner

    def __call__(self, scope):
        return WsTokenAuthMiddlewareInstance(scope, self)

class OnTheFlyJWTAuthentication(BaseJSONWebTokenAuthentication):

    @classmethod
    def decode_token(self, raw_token, verify=False):
        """
        Validates an encoded JSON web token and returns a validated token
        wrapper object.
        """
        messages = []

        for AuthToken in api_settings.AUTH_TOKEN_CLASSES:
            try:
                return AuthToken(raw_token, verify)
            except TokenError as e:
                messages.append({'token_class': AuthToken.__name__,
                                 'token_type': AuthToken.token_type,
                                 'message': e.args[0]})

        raise InvalidToken({
            'detail': _('Given token not valid for any token type'),
            'messages': messages,
        })

    def authenticate_token(self, raw_token):
        validated_token = self.get_validated_token(raw_token)
        return self.get_user(validated_token), validated_token

class WsTokenAuthMiddlewareInstance(OnTheFlyJWTAuthentication):
    """
    Token authorization middleware for Django Channels 2
    """

    def get_query_params(self, scope):
        return parse_qs(scope['query_string'].decode())

    def __init__(self, scope, middleware):
        self.middleware = middleware
        self.scope = dict(scope)
        self.inner = self.middleware.inner
        super(WsTokenAuthMiddlewareInstance, self).__init__()

    async def __call__(self, receive, send):
        close_old_connections()
        query_params = self.get_query_params(self.scope)
        if "auth_token" in query_params:
            try:
                auth_token = query_params['auth_token'][0]
                user, access_token = self.authenticate_token(auth_token)
                self.scope['user'] = user
            except Exception as e:
                traceback.print_exc()
                self.scope['user'] = AnonymousUser()
        inner = self.inner(self.scope)
        return await inner(receive, send)

WsTokenAuthMiddlewareStack = lambda inner: WsTokenAuthMiddleware(AuthMiddlewareStack(inner))


from django.conf import settings
from django.contrib.admin import ACTION_CHECKBOX_NAME
from django.contrib.auth import get_user_model


def is_anonymous_user(user):
    if  user is None or user.is_anonymous:
        return True
    User = get_user_model()
    username = getattr(user, User.USERNAME_FIELD)
    if hasattr(settings, 'ANONYMOUS_USER_NAME'):
        anon_username = settings.ANONYMOUS_USER_NAME

    else:
        from guardian.conf import settings as guardian_settings
        anon_username = guardian_settings.ANONYMOUS_USER_NAME
    return anon_username and anon_username == username


def changelist_no_args(action_name, model):
    def decorator(function):
        def wrapper(self, request, *args, **kwargs):
            if 'action' in request.POST:
                if (type(action_name) == list and request.POST['action'] in action_name) or request.POST['action'] == action_name:
                    if not request.POST.getlist(ACTION_CHECKBOX_NAME):
                        post = request.POST.copy()
                        for u in model.objects.all():
                            post.update({ACTION_CHECKBOX_NAME: str(u.id)})
                        request._set_post(post)
            function(self, request, *args, **kwargs)
            return super(self.__class__, self).changelist_view(request, *args, **kwargs)
        return wrapper
    return decorator


from rest_framework.parsers import BaseParser


class PlainTextParser(BaseParser):
    """
    Plain text parser.
    """
    media_type = 'text/plain'

    def parse(self, stream, media_type=None, parser_context=None):
        """
        Simply return a string representing the body of the request.
        """
        return stream.read()